/* lcs2d3.c ***********************************************\
*                                                          *
*  Higher order Godunov-type scheme for 2D Euler equations *
*                   on Cartesian grid                      *
*  July 7, 2001                                            *
*  author: Andrei A. Chernousov                            *
*          E-mail: <andrei99@iname.com>                    *
*          http://www.geocities.com/andrei_chernousov      *
\**********************************************************/

/* 
Features:
  piecewise-parabolic spatial approximation of solution within cell
  linearized characteristic-based approximate Riemann solver (LCS)
  three step Runge-Kutta advancing in time
  optimized for speed
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include <signal.h>

/* changed by SIGINT handler*/
int continFlag = 1;

/* SIGINT handler */
#pragma argsused
void termHandler(int sigType)
{
  if (continFlag--)  /* 1 -> 0 */
	;
  printf("[Ctrl+C]: SIGINT caught, terminating.\n");
}


#define Cv      717.75 /* air heat capacity at constant volume, J/(kg*K) */
#define R_VOZD  287.10 /* gas constant       */
#define KR_VOZD 401.94 /* gas constant * 1.4 */

unsigned
  LEN,   /* number of 2D cells in x-direction */
  HIG,   /* number of 2D cells in y-direction */
  LENN,  /* LEN+1 */
  HIGG;  /* HIG+1 */

float
  **U1,  **U2,  **U3,  **U4,  /* arrays for conservative variables */
  **U1p, **U2p, **U3p, **U4p, /* - the final or intermedial states */

  **U1_, **U2_, **U3_, **U4_, /* pointers for substitution */

  **xU1, **xU2, **xU3, **xU4, /* arrays at cell x-boundaries */
  **U1x, **U2x, **U3x, **U4x,

  **yU1, **yU2, **yU3, **yU4, /* arrays at cell y-boundaries */
  **U1y, **U2y, **U3y, **U4y,

  u1, u2, u3, u4;

  /* primitive variables */
float 
  R, U, V, P, C, R_, U_, V_, P_, C_, _R, _U, _V, _P, _C;

  /* for LCS procedure */
float 
  C_p, C_m, C_o,
  _jo, _jp, _jm, jo_, jp_, jm_,
  al_p, al_m, al_o,
  J_p, J_m, J_o;

  /* stuff */
unsigned long step, numstep, mem;
char Stage, Answer;

  /* grid spacings and steps */
float deltaX, deltaY, deltaT, deltaT_X, deltaT_Y;

  /* poiter to file descriptor */
FILE *pF;

  /* alias names for fluxes */
  /* x */
#define Fx1 xU1  /* mass flux along X ( RU )          */
#define Fx2 xU2  /* x-momentum flux along X ( RUU+P ) */
#define Fx3 xU3  /* y-momentum flux along X ( RUV )   */
#define Fx4 xU4  /* energy flux along X ( U(RE+P) )   */
  /* y */
#define Fy1 yU1  /* mass flux along Y ( RV )          */
#define Fy2 yU2  /* x-momentum flux along Y ( RVU )   */
#define Fy3 yU3  /* y-momentum flux along Y ( RVV+P ) */
#define Fy4 yU4  /* energy flux along Y ( V(RE+P) )   */

  /* expressions with the ratio of specific heats */
float
	K,        // 1.4
	K_1,      // 0.4
	_K_1,     // (-0.4)
	_05K,     // 0.7
	K_K_1,    // 3.5
	K_1_2,    // 0.2
	_1_2_K_1; // 1.25


/************
*  ARRAY2D  *   allocates memory for 2D ararys
*************/
float **Array2D(unsigned columns, unsigned rows)
{
float **x;
unsigned i;

  if ((x = (float **)malloc(columns * sizeof(float *))) == NULL) {
	fprintf(stderr, "cannot allocate memory\n");
	exit(-1);
  }
  for (i = 0; i < columns; i++)
	if ((x[i] = (float *)malloc(rows * sizeof(float))) == NULL) {
	  fprintf(stderr, "cannot allocate memory\n");
	  exit(-1);
  }
  return x;

} /* end Array2D() */


/***************
*  INITIALIZE  *   performs initializations
***************/
void Initialize(void)
{
unsigned i, j;
char str[80];

  /* reading inits from the file*/
  if ((pF = fopen("lcs2d3.ini", "r")) == NULL) {
	fprintf(stderr, "cannot open file \"lcs2d3.ini\"\n");
	exit(-1);
  }
  i = 0;
  do {
  	if (fgets(str, 79, pF) != NULL) {
	  switch (i) {
	  case 0:	 LEN   = atoi(str); break;
	  case 1:	 HIG   = atoi(str); break;
	  case 2:  deltaX = atof(str); break;
	  case 3:  deltaY = atof(str); break;
	  case 4:  deltaT = atof(str); break;
	  case 5:  Answer = atoi(str); break;
	  case 6:  numstep= atol(str); break;
	  default: break;
	  } /* end switch */
	  i++;
	} /* end if()  */
	else {
	  fprintf(stderr, "error reading file \"lsc2d3.ini\"\n");
	  exit(-1);
	}
  } while (i < 7);
  fclose(pF);

  /* some other initializations */
  deltaT_X = deltaT / deltaX;
  deltaT_Y = deltaT / deltaY;
  LENN = LEN+1;
  HIGG = HIG+1;

  /* expressions with the ratio of specific heats */
  K = 1.4;
  K_1 = K - 1.;
  _K_1 = - K_1;
  _05K = 0.5 * K;
  K_K_1 = K / K_1;
  K_1_2 = K_1 / 2;
  _1_2_K_1 = 1. / (K_1 + K_1);

  /* approx. amount of dynamic memory needed */
  mem = sizeof(float) * 4 * 
  ( ((long)(LEN+2) * (long)(HIG+2))
+ ( (long)( LEN+1 ) * (long)  HIG     ) * 2
+ ( (long)  LEN     * (long)( HIG+1 ) ) * 2 );
  printf("%lu bytes of dynamic memory required. Allocating...", mem);
  /* conservative variables in cells */
  U1 = Array2D(LEN+2, HIG+2);
  U2 = Array2D(LEN+2, HIG+2);
  U3 = Array2D(LEN+2, HIG+2);
  U4 = Array2D(LEN+2, HIG+2);
  U1p= Array2D(LEN+2, HIG+2);
  U2p= Array2D(LEN+2, HIG+2);
  U3p= Array2D(LEN+2, HIG+2);
  U4p= Array2D(LEN+2, HIG+2);
	/* y-boundaries */
  xU1 = Array2D(LEN+1, HIG);
  xU2 = Array2D(LEN+1, HIG);
  xU3 = Array2D(LEN+1, HIG);
  xU4 = Array2D(LEN+1, HIG);
  U1x = Array2D(LEN+1, HIG);
  U2x = Array2D(LEN+1, HIG);
  U3x = Array2D(LEN+1, HIG);
  U4x = Array2D(LEN+1, HIG);
	/* y-boundaries */
  yU1 = Array2D(LEN, HIG+1);
  yU2 = Array2D(LEN, HIG+1);
  yU3 = Array2D(LEN, HIG+1);
  yU4 = Array2D(LEN, HIG+1);
  U1y = Array2D(LEN, HIG+1);
  U2y = Array2D(LEN, HIG+1);
  U3y = Array2D(LEN, HIG+1);
  U4y = Array2D(LEN, HIG+1);
  printf(" done!\n");

  if (Answer) {
	if ((pF = fopen("backup", "rb")) == NULL) {
	  fprintf(stderr, "cannot open file \"backup\"\n");
	  exit(-1);
	}
	for (i = 1; i <= LEN; i++) {
	  fread(&U1[i][1], HIG * sizeof(float), 1, pF);
	  fread(&U2[i][1], HIG * sizeof(float), 1, pF);
	  fread(&U3[i][1], HIG * sizeof(float), 1, pF);
	  fread(&U4[i][1], HIG * sizeof(float), 1, pF);
	}
	fread(&step, sizeof( unsigned long ), 1, pF);
	fclose(pF);
  }
  else {
	V = U = 0.;
	for (i = 1; i <= LEN; i++) {
	  for (j = 1; j <= HIG; j++) {
		if (i <= LEN/2 && j <= HIG/2) {
		  P = 200000.;
		  R = 2.0;
		}
		else {
		  P = 100000.;
		  R = 1.0;
		}
		U1[i][j] = R;
		U2[i][j] = R * U;
		U3[i][j] = R * V;
		U4[i][j] = P / K_1 + 0.5 * R * (U * U + V * V);
	  }
	}
	step = 0;
  } /* end else() */

  Stage = 1; U1_ = U1, U2_ = U2, U3_ = U3, U4_ = U4;

} /* end Initialize() */


/********************
* BOUNDARYCONDITONS *   boundary conditions
********************/
void BoundaryConditions(void)
{
register unsigned i, j;

  /* upper and lower boundaries */
  for (i = 1; i < LENN; i++) {
	U1_[i][HIGG] =   U1_[i][HIG];
	U2_[i][HIGG] =   U2_[i][HIG];
	U3_[i][HIGG] = - U3_[i][HIG];
	U4_[i][HIGG] =   U4_[i][HIG];
	U1_[i][0] =   U1_[i][1];
	U2_[i][0] =   U2_[i][1];
	U3_[i][0] = - U3_[i][1];
	U4_[i][0] =   U4_[i][1];
  } /* end for() */
  
  /* left and right boundaries */
  for (j = 1; j < HIGG; j++) {
	U1_[LENN][j] =   U1_[LEN][j];
	U2_[LENN][j] = - U2_[LEN][j];
	U3_[LENN][j] =   U3_[LEN][j];
	U4_[LENN][j] =   U4_[LEN][j];
	U1_[0][j] =   U1_[1][j];
	U2_[0][j] = - U2_[1][j];
	U3_[0][j] =   U3_[1][j];
	U4_[0][j] =   U4_[1][j];
  } /* end for() */

} /* end BounaryConditions( ) */


/*--- For monotone piecevise parabolic reconstruction ---*/

  /* constant coefficients: 4.0, 1/3, 1/6 are recommended */
#define C1 0.333333333
#define C2 0.166666667
#define BB 4.000000000

/***************
* minmod(x, y) * - macro: limiting function of Chakravarthy, Osher
***************/
/*
	- uses register variable kk defined in Reconstruction()
	- don't use expressions as arguments of this macro!
	- beware of other side effects
*/
#define minmod(x, y) ( (x*y <= 0.) ? 0.: \
  ( (kk=(y>0.)?y*BB:-y*BB, x>0.) ? ((x<kk)?x:kk) : ((-x<kk)?x:-kk) ) )

/***************
* minmod(x, y) * - function: limiting function of Chakravarthy, Osher (slow!)
***************/
/*
float ay;
float minmod(float x, float y)
{
  if (x * y <= 0.)
	return 0.;

  ay = ((y *= BB) > 0.) ? y: -y;

  if (x > 0.)
	return (x < ay) ? x : ay;

  return (-x < ay) ? x: -ay;

} */ /* end minmod() */

float
  /* finite differences of conservative variables */
  m1, m2, m3, m4, w1, w2, w3, w4,
  /* characteristic differences */
  M1, M2, M3, M4, W1, W2, W3, W4,
  /* modified characteristic differences */
  _M1, _M2, _M3, _M4, _W1, _W2, _W3, _W4,
  /* transformation matrix */
  S11, S12, S13,
  S21,
  S31, S32, S33,
  S41, S42, S43,
  /* inverted transformation matrix */
  				_S13, _S14,
				_S23, _S24,
				_S33, _S34,
		  _S42, _S43, _S44,
  /* common expresions */
  UU_VV,
  K_1_2_UU_VV,
  UU_VV_4CC,
  _UU_VV_2CC,
  CC, _1_CC, _U_CC, _V_CC, _1_2C, _1_2CC,
  _K_1_U, _K_1_V,
  U_1_2CC,	V_1_2CC;

/*******************
*  RECONSTRUCTION  *   piecewise-parabolic reconstruction of the cell parameters with 2..3 order in space
*******************/
void Reconstruction(void)
{
register float kk, k1, k2, k3, k4;
register unsigned _i, _j, i, j, i_, j_;


/* limited piecewise-parabolic reconstruction */
for (_i = 0, i = 1, i_ = 2; i < LENN; _i++, i++, i_++) {
  for (_j = 0, j = 1, j_ = 2; j < HIGG; _j++, j++, j_++) {
	/* at cell centroid */
	U =  (u2 = U2_[i][j]) / (R = u1 = U1_[i][j]);
	V =  (u3 = U3_[i][j]) /  R;
	P = ((u4 = U4_[i][j]) - 0.5 * R * (U * U + V * V)) * K_1;
	C = sqrt(K * P / R);
	/* common expresions */
	CC = C * C;
	UU_VV = U * U + V * V;
	K_1_2_UU_VV = K_1_2 * UU_VV;
	UU_VV_4CC = UU_VV / (4. * CC);
	_UU_VV_2CC = -(UU_VV_4CC + UU_VV_4CC);
	/**/
	_1_CC = - 1 / CC;
	_U_CC = - U / CC;
	_V_CC = - V / CC;
	/**/
	_1_2C = 1./(C+C);
	_1_2CC = 1./(CC + CC);
	/**/
	_K_1_U = _K_1 * U;
	_K_1_V = _K_1 * V;
	/**/
	U_1_2CC = U * _1_2CC;
	V_1_2CC = V * _1_2CC;

	/*--- X ---*/
	/* transformation matrix */
	S11 = -C*C + K_1_2_UU_VV; S12 =   _K_1_U; //S13 = _K_1_V;  S14 = K_1;
	S21 = -V;               //S22 =       0.; //S23 =     1.;  S24 =  0.;
	S31 = -C*U + K_1_2_UU_VV; S32 = C+_K_1_U; //S33 = _K_1_V;  S34 = K_1;
	S41 =  C*U + K_1_2_UU_VV; S42 =-C+_K_1_U; //S43 = _K_1_V;  S44 = K_1;
	/* inverted transformation matrix */
	/*_S11 = _1_CC;*/   /*_S12 = 0.;*/ _S13 =                                 _S14 = _1_2CC;
	/*_S21 = _U_CC;*/   /*_S22 = 0.;*/ _S23 = U_1_2CC +_1_2C;                 _S24 = U_1_2CC - _1_2C;
	/*_S31 = _V_CC;*/   /*_S32 = 1.;*/ _S33 =                                 _S34 = V_1_2CC;
	/*_S41 = _UU_VV_2CC;*/_S42 =  V;   _S43 = UU_VV_4CC + U*_1_2C + _1_2_K_1; _S44 = UU_VV_4CC - U*_1_2C + _1_2_K_1;
	/* differences */
	/* forward */
	m1 = U1_[i_][j] - u1;
	m2 = U2_[i_][j] - u2;
	m3 = U3_[i_][j] - u3;
	m4 = U4_[i_][j] - u4;
	/* backward */
	w1 = u1 - U1_[_i][j];
	w2 = u2 - U2_[_i][j];
	w3 = u3 - U3_[_i][j];
	w4 = u4 - U4_[_i][j];
	/* differences of characteristic fields */
	/* forward */
	M1 = S11*m1 + S12*m2 + _K_1_V*m3 + K_1*m4;
	M2 = S21*m1 +                 m3         ;
	M3 = S31*m1 + S32*m2 + _K_1_V*m3 + K_1*m4;
	M4 = S41*m1 + S42*m2 + _K_1_V*m3 + K_1*m4;
	/* backward */
	W1 = S11*w1 + S12*w2 + _K_1_V*w3 + K_1*w4;
	W2 = S21*w1 +                 w3         ;
	W3 = S31*w1 + S32*w2 + _K_1_V*w3 + K_1*w4;
	W4 = S41*w1 + S42*w2 + _K_1_V*w3 + K_1*w4;
	/* modified characteristic differences */
	_M1 = minmod(M1, W1);
	_M2 = minmod(M2, W2);
	_M3 = minmod(M3, W3);
	_M4 = minmod(M4, W4);
	_W1 = minmod(W1, M1);
	_W2 = minmod(W2, M2);
	_W3 = minmod(W3, M3);
	_W4 = minmod(W4, M4);
	/* at rightmost interface */
	k1 = _M1 * C1 + _W1 * C2;
	k2 = _M2 * C1 + _W2 * C2;
	k3 = _M3 * C1 + _W3 * C2;
	k4 = _M4 * C1 + _W4 * C2;
	xU1[ i][_j] = u1 +      _1_CC * k1             + _S13 * k3 + _S14 * k4;
	xU2[ i][_j] = u2 +      _U_CC * k1             + _S23 * k3 + _S24 * k4;
	xU3[ i][_j] = u3 +      _V_CC * k1 +        k2 + _S33 * k3 + _S34 * k4;
	xU4[ i][_j] = u4 + _UU_VV_2CC * k1 + _S42 * k2 + _S43 * k3 + _S44 * k4;
	/* at leftmost interface */
	k1 = _M1 * C2 + _W1 * C1;
	k2 = _M2 * C2 + _W2 * C1;
	k3 = _M3 * C2 + _W3 * C1;
	k4 = _M4 * C2 + _W4 * C1;
	U1x[_i][_j] = u1 -      _1_CC * k1             - _S13 * k3 - _S14 * k4;
	U2x[_i][_j] = u2 -      _U_CC * k1             - _S23 * k3 - _S24 * k4;
	U3x[_i][_j] = u3 -      _V_CC * k1 -        k2 - _S33 * k3 - _S34 * k4;
	U4x[_i][_j] = u4 - _UU_VV_2CC * k1 - _S42 * k2 - _S43 * k3 - _S44 * k4;
	/*--- Y ---*/
	/* transformation matrix */
	S11 = -C*C + K_1_2_UU_VV; S12 = /*_K_1_U;*/ S13 =   _K_1_V; //S14 = K_1;
	S21 = /*U;*/            //S22 =      -1.;   S23 =       0.;   S24 =  0.;
	S31 = -C*V + K_1_2_UU_VV; S32 = /*_K_1_U;*/ S33 = C+_K_1_V; //S34 = K_1;
	S41 =  C*V + K_1_2_UU_VV; S42 = /*_K_1_U;*/ S43 =-C+_K_1_V; //S44 = K_1;
	/* inverted transformation matrix */
	/*_S11 = _1_CC;*/   /*_S12 =  0.;*/ _S13 =                                   _S14 = _1_2CC;
	/*_S21 = _U_CC;*/   /*_S22 = -1.;*/ _S23 =                                   _S24 = U_1_2CC;
	/*_S31 = _V_CC;*/   /*_S32 =  0.;*/ _S33 = V_1_2CC + _1_2C;				       _S34 = V_1_2CC - _1_2C;
	/*_S41 = _UU_VV_2CC;*/_S42 = -U;    _S43 = UU_VV_4CC + V * _1_2C + _1_2_K_1; _S44 = UU_VV_4CC - V * _1_2C + _1_2_K_1;
	/* differences */
	/* upper */
	m1 = U1_[i][j_] - u1;
	m2 = U2_[i][j_] - u2;
	m3 = U3_[i][j_] - u3;
	m4 = U4_[i][j_] - u4;
	/* lower */
	w1 = u1 - U1_[i][_j];
	w2 = u2 - U2_[i][_j];
	w3 = u3 - U3_[i][_j];
	w4 = u4 - U4_[i][_j];
	/* differences of characteristic fields */
	/* upper */
	M1 = S11*m1 + _K_1_U*m2 + S13*m3 + K_1*m4;
	M2 =   U*m1 -        m2                  ;
	M3 = S31*m1 + _K_1_U*m2 + S33*m3 + K_1*m4;
	M4 = S41*m1 + _K_1_U*m2 + S43*m3 + K_1*m4;
	/* lower */
	W1 = S11*w1 + _K_1_U*w2 + S13*w3 + K_1*w4;
	W2 =   U*w1 -        w2                  ;
	W3 = S31*w1 + _K_1_U*w2 + S33*w3 + K_1*w4;
	W4 = S41*w1 + _K_1_U*w2 + S43*w3 + K_1*w4;
	/* limited */
	_M1 = minmod(M1, W1);
	_M2 = minmod(M2, W2);
	_M3 = minmod(M3, W3);
	_M4 = minmod(M4, W4);
	_W1 = minmod(W1, M1);
	_W2 = minmod(W2, M2);
	_W3 = minmod(W3, M3);
	_W4 = minmod(W4, M4);
	/* at upper interface */
	k1 = _M1 * C1 + _W1 * C2;
	k2 = _M2 * C1 + _W2 * C2;
	k3 = _M3 * C1 + _W3 * C2;
	k4 = _M4 * C1 + _W4 * C2;
	yU1[_i][ j] = u1 +      _1_CC * k1             + _S13 * k3 + _S14 * k4;
	yU2[_i][ j] = u2 +      _U_CC * k1 -        k2 + _S23 * k3 + _S24 * k4;
	yU3[_i][ j] = u3 +      _V_CC * k1             + _S33 * k3 + _S34 * k4;
	yU4[_i][ j] = u4 + _UU_VV_2CC * k1 + _S42 * k2 + _S43 * k3 + _S44 * k4;
	/* at lower interface */
	k1 = _M1 * C2 + _W1 * C1;
	k2 = _M2 * C2 + _W2 * C1;
	k3 = _M3 * C2 + _W3 * C1;
	k4 = _M4 * C2 + _W4 * C1;
	U1y[_i][_j] = u1 -      _1_CC * k1             - _S13 * k3 - _S14 * k4;
	U2y[_i][_j] = u2 -      _U_CC * k1 +        k2 - _S23 * k3 - _S24 * k4;
	U3y[_i][_j] = u3 -      _V_CC * k1             - _S33 * k3 - _S34 * k4;
	U4y[_i][_j] = u4 - _UU_VV_2CC * k1 - _S42 * k2 - _S43 * k3 - _S44 * k4;
  }
}

} /* end Reconstruction() */


/*******************
*  BConInterfaces  *   completion of the solution at the outer sides of cell boundaries
*******************/
void BConInterfaces( void )
{
register unsigned i, j;

  /* upper and lower boundaries */
  for( i = 0; i < LEN; i++ ) {
	yU1[i][0] =   U1y[i][0];
	yU2[i][0] =   U2y[i][0];
	yU3[i][0] = - U3y[i][0];
	yU4[i][0] =   U4y[i][0];
	U1y[i][HIG] =   yU1[i][HIG];
	U2y[i][HIG] =   yU2[i][HIG];
	U3y[i][HIG] = - yU3[i][HIG];
	U4y[i][HIG] =   yU4[i][HIG];
  }

  /* left and right boundaties */
  for( j = 0; j < HIG; j++ ) {
	xU1[0][j] =   U1x[0][j];
	xU2[0][j] = - U2x[0][j];
	xU3[0][j] =   U3x[0][j];
	xU4[0][j] =   U4x[0][j];
	U1x[LEN][j] =   xU1[LEN][j];
	U2x[LEN][j] = - xU2[LEN][j];
	U3x[LEN][j] =   xU3[LEN][j];
	U4x[LEN][j] =   xU4[LEN][j];
  }

} /* end BConInterfaces( ) */


/***********
*  FLUXES  *   fluxes at the boundaries
***********/
void Fluxes(void)
{
register unsigned i, j;
register float RU; /* convective normal mass flux */

  /*--- X-direction ---*/
  for (i = 0; i < LENN; i++) {
	for (j = 0; j < HIG; j++) {
	  /* left */
	  _U =  xU2[i][j] / (_R = xU1[i][j]);
	  _V =  xU3[i][j] /  _R;
	  _P = (xU4[i][j] - 0.5 * _R * (_U *_U + _V *_V)) * K_1;
	  _C = sqrt(K * _P / _R);
	  _jo = _K_1 * _P; /*_P - _C * _C * _R; */
	  _jp = _U + (RU = _P / (_R * _C));
	  _jm = _U -  RU;
	  /* right */
	  U_ =  U2x[i][j] / (R_ = U1x[i][j]);
	  V_ =  U3x[i][j] /  R_;
	  P_ = (U4x[i][j] - 0.5 * R_ * (U_* U_ + V_* V_)) * K_1;
	  C_ = sqrt(K * P_ / R_);
	  jo_ = _K_1 * P_; /*P_ - C_ * C_ * R_;*/
	  jp_ = U_ + (RU = P_ / (R_ * C_));
	  jm_ = U_ -  RU;
	  /* linearized procedure */
	  U = 0.5 * (_U + U_);
	  C = 0.5 * (_C + C_);
	  C_p = U + C; C_m = U - C; C_o = U;
	  /* J_p and al_p */
	  if (C_p > 0.) { al_p = K_1_2 * (_jm - _jp) / _jo; J_p  = _jp; }
	  else          { al_p = K_1_2 * (jm_ - jp_) / jo_; J_p  = jp_; }
	  /* J_m and al_m */
	  if (C_m > 0.) { al_m = K_1_2 * (_jp - _jm) / _jo; J_m  = _jm; }
	  else          { al_m = K_1_2 * (jp_ - jm_) / jo_; J_m  = jm_; }
	  /* J_o and al_o */
	  if (C_o > 0.) { al_o = _05K * (_jp - _jm); al_o = - al_o * al_o; J_o  = _jo; }
	  else          { al_o = _05K * (jp_ - jm_); al_o = - al_o * al_o; J_o  = jo_; }
	  /* parameters on the interface */
	  P = (J_p - J_m) / (al_p - al_m);
	  U = J_p - al_p * P;
	  R = (J_o - P) / al_o;
	  C = sqrt(K * P / R);
	  /* tangential component of velocity */
	  V = (U > 0.) ? _V: V_;
	  /* flux vector */
	  Fx1[i][j] = RU = R * U;
	  Fx2[i][j] = RU * U + P;
	  Fx3[i][j] = RU * V;
	  Fx4[i][j] = RU * (K_K_1 * P / R + 0.5 * (U * U + V * V));
	} /* end for */
  } /* end for */

  /*--- Y-direction ---*/
  for (j = 0; j < HIGG; j++) {
	for (i = 0; i < LEN; i++) {
	  /* lower */
	  _U =  yU2[i][j] / (_R = yU1[i][j]);
	  _V =  yU3[i][j] /  _R;
	  _P = (yU4[i][j] - 0.5 * _R * (_U *_U + _V *_V)) * K_1;
	  _C = sqrt(K * _P / _R);
	  _jo = _K_1 * _P; /*_P - _C * _C * _R;*/
	  _jp = _V + (RU = _P / (_R * _C));
	  _jm = _V -  RU;
	  /* upper */
	  U_ =  U2y[i][j] / (R_ = U1y[i][j]);
	  V_ =  U3y[i][j] /  R_;
	  P_ = (U4y[i][j] - 0.5 * R_ * (U_* U_ + V_* V_)) * K_1;
	  C_ = sqrt(K * P_ / R_);
	  jo_ = _K_1 * P_; /*P_ - C_ * C_ * R_;*/
	  jp_ = V_ + (RU = P_ / (R_ * C_));
	  jm_ = V_ -   RU;
	  /* linearized procedure */
	  V = 0.5 * (_V + V_);
	  C = 0.5 * (_C + C_);
	  C_p = V + C; C_m = V - C; C_o = V;
	  /* Jp and al_p */
	  if (C_p > 0.) { al_p = K_1_2 * (_jm - _jp) / _jo; J_p  = _jp; }
	  else          { al_p = K_1_2 * (jm_ - jp_) / jo_; J_p  = jp_; }
	  /* Jm and al_m */
	  if (C_m > 0.) { al_m = K_1_2 * (_jp - _jm) / _jo; J_m  = _jm; }
	  else          { al_m = K_1_2 * (jp_ - jm_) / jo_; J_m  = jm_; }
	  /* Jo and al_o */
	  if (C_o > 0.) { al_o = _05K * (_jp - _jm); al_o = - al_o * al_o; J_o  = _jo; }
	  else          { al_o = _05K * (jp_ - jm_); al_o = - al_o * al_o; J_o  = jo_; }
	  /* parameters on the interface */
	  P = (J_p - J_m) / (al_p - al_m);
	  V = J_p - al_p * P;
	  R = (J_o - P) / al_o;
	  C = sqrt(K * P / R);
	  /* tangential component of velocitiy */
	  U = (V > 0.) ? _U: U_;
	  /* fluxes */
	  Fy1[i][j] = RU = R * V;
	  Fy2[i][j] = RU * U;
	  Fy3[i][j] = RU * V + P;
	  Fy4[i][j] = RU * (K_K_1 * P / R + 0.5 * (U * U + V * V));
	} /* end for */
  } /* end for */

} /* end Fluxes() */


/**************
*  EVOLUTION  *   new (or intermediate) values of conservative variables
**************/
void Evolution(void)
{
register unsigned i, j, _i, _j;

  switch (Stage) {
  
	case 1: /* 1st stage */
	  for (i = 1, _i = 0; i < LENN; i++, _i++) {
		for (j = 1, _j = 0; j < HIGG; j++, _j++) {
		  U1p[i][j] = U1[i][j] + deltaT_X * (Fx1[_i][_j] - Fx1[i][_j]) + deltaT_Y * (Fy1[_i][_j] - Fy1[_i][j]);
		  U2p[i][j] = U2[i][j] + deltaT_X * (Fx2[_i][_j] - Fx2[i][_j]) + deltaT_Y * (Fy2[_i][_j] - Fy2[_i][j]);
		  U3p[i][j] = U3[i][j] + deltaT_X * (Fx3[_i][_j] - Fx3[i][_j]) + deltaT_Y * (Fy3[_i][_j] - Fy3[_i][j]);
		  U4p[i][j] = U4[i][j] + deltaT_X * (Fx4[_i][_j] - Fx4[i][_j]) + deltaT_Y * (Fy4[_i][_j] - Fy4[_i][j]);
		}
	  }
	  /* next stage */
	  Stage = 2; U1_ = U1p, U2_ = U2p, U3_ = U3p, U4_ = U4p;
	  break;

	case 2: /* 2nd stage */
	  for (i = 1, _i = 0; i < LENN; i++, _i++) {
		for (j = 1, _j = 0; j < HIGG; j++, _j++) {
		  U1p[i][j] = 0.75 * U1[i][j] + 0.25 * ( U1p[i][j] + deltaT_X * ( Fx1[_i][_j] - Fx1[i][_j] ) + deltaT_Y * ( Fy1[_i][_j] - Fy1[_i][j] ) );
		  U2p[i][j] = 0.75 * U2[i][j] + 0.25 * ( U2p[i][j] + deltaT_X * ( Fx2[_i][_j] - Fx2[i][_j] ) + deltaT_Y * ( Fy2[_i][_j] - Fy2[_i][j] ) );
		  U3p[i][j] = 0.75 * U3[i][j] + 0.25 * ( U3p[i][j] + deltaT_X * ( Fx3[_i][_j] - Fx3[i][_j] ) + deltaT_Y * ( Fy3[_i][_j] - Fy3[_i][j] ) );
		  U4p[i][j] = 0.75 * U4[i][j] + 0.25 * ( U4p[i][j] + deltaT_X * ( Fx4[_i][_j] - Fx4[i][_j] ) + deltaT_Y * ( Fy4[_i][_j] - Fy4[_i][j] ) );
		}
	  }
	  /* next stage */
	  Stage = 3;
	  break;

	default: /* 3rd stage */
	  for (i = 1, _i = 0; i < LENN; i++, _i++) {
		for (j = 1, _j = 0; j < HIGG; j++, _j++) {
		  U1[i][j] = 0.33333333333 * U1[i][j] + 0.66666666667 * (U1p[i][j] + deltaT_X * (Fx1[_i][_j] - Fx1[i][_j]) + deltaT_Y * (Fy1[_i][_j] - Fy1[_i][j]));
		  U2[i][j] = 0.33333333333 * U2[i][j] + 0.66666666667 * (U2p[i][j] + deltaT_X * (Fx2[_i][_j] - Fx2[i][_j]) + deltaT_Y * (Fy2[_i][_j] - Fy2[_i][j]));
		  U3[i][j] = 0.33333333333 * U3[i][j] + 0.66666666667 * (U3p[i][j] + deltaT_X * (Fx3[_i][_j] - Fx3[i][_j]) + deltaT_Y * (Fy3[_i][_j] - Fy3[_i][j]));
		  U4[i][j] = 0.33333333333 * U4[i][j] + 0.66666666667 * (U4p[i][j] + deltaT_X * (Fx4[_i][_j] - Fx4[i][_j]) + deltaT_Y * (Fy4[_i][_j] - Fy4[_i][j]));
		}
	  }
	  /* next stage */
	  Stage = 1; U1_ = U1, U2_ = U2, U3_ = U3, U4_ = U4;
	  break;

	} /* end switch() */

} /* end Evolution() */


/*********
*  SAVE  *   backup copy of solution in binary file
*********/
void Save(void)
{
unsigned i, j;

  if ((pF = fopen("backup", "wb")) == NULL) {
	fprintf(stderr, "cannot open file \"backup\"\n");
	exit(-1);
  }

  for (i = 1; i <= LEN; i++) {
	fwrite(&U1[i][1], HIG * sizeof(float), 1, pF);
	fwrite(&U2[i][1], HIG * sizeof(float), 1, pF);
	fwrite(&U3[i][1], HIG * sizeof(float), 1, pF);
	fwrite(&U4[i][1], HIG * sizeof(float), 1, pF);
  }
  fwrite(&step, sizeof(unsigned long), 1, pF);

  fclose(pF);

} /* end Save() */


/***********
*  OUTPUT  *   output of the flowfield in format suitable for visualisation with WinField utility
***********/
void Output(void)
{
unsigned i, j;
char str[80];

  sprintf(str, "%lu", step);
  if ((pF = fopen(str, "w")) == NULL) {
	fprintf(stderr, "cannot open file \"%s\"\n", str);
	exit(-1);
  }

  fprintf(pF, "%d\n%d\n%f\n%f\n", LEN, HIG, deltaX, deltaY);
  for (j = 1; j < HIGG; j++)
  	for (i = 1; i < LENN; i++)
	  fprintf(pF, "%1.3f \n", U1[i][j]);
  fclose(pF);

} /* end Output() */


/*********
*  MAIN  *    main function
*********/
int main(void)
{
  /*--- initializaton ---*/
  Initialize();

  /*--- handler for a SIGINT signal ---*/
  signal(SIGINT, termHandler);

  /*--- time integration loop ---*/
  while (1) {

	/*--- exit after _numstep_ steps ---*/
	if (step == numstep || !continFlag)
	  break;
	printf( "%lu\n", ++step );
//	step++;

	/*=== 1st stage ===*/
	/*--- boundary conditions ---*/
	BoundaryConditions();
	/*--- piecewise-parabolic reconstruction ---*/
	Reconstruction();
	/*--- at the outer sides of cell boundaries ---*/
	BConInterfaces();
	/*--- fluxes at the boundaries ---*/
	Fluxes();
	/*--- intermediate values of conservative variables ---*/
	Evolution();

	/*=== 2nd stage ===*/
	/*--- boundary conditions ---*/
	BoundaryConditions();
	/*--- piecewise-parabolic reconstruction ---*/
	Reconstruction();
	/*--- at the outer sides of cell boundaries ---*/
	BConInterfaces();
	/*--- fluxes at the boundaries ---*/
	Fluxes();
	/*--- intermediate values of conservative variables ---*/
	Evolution();

	/*=== 3rd stage ===*/
	/*--- boundary conditions ---*/
	BoundaryConditions();
	/*--- piecewise-parabolic reconstruction ---*/
	Reconstruction();
	/*--- at the outer sides of cell boundaries ---*/
	BConInterfaces();
	/*--- fluxes at the boundaries ---*/
	Fluxes();
	/*--- new values of conservative variables ---*/
	Evolution();

  } /* end while() */

  printf("%f %f\n", U1[35][25], U1[25][35]);

  /*--- backup copy of solution in binary file ---*/
  Save();

  /*--- output of the flowfield ---*/
  Output();

  return 0;

} /* end main() */

/* end lcs2d3.c */

